import enum
import json
import os
import shutil
from collections import deque
from pathlib import Path
from typing import cast

import attr
import cattrs
from rich import print
from rich.progress import Progress
from tomlkit import load as toml_load

from kamuidrome.cache import ModCache
from kamuidrome.meta import LocalMetadata, PackMetadata
from kamuidrome.modrinth.client import ModrinthApi
from kamuidrome.modrinth.models import ModSideValue, ProjectId, VersionId
from kamuidrome.modrinth.utils import VersionResult
from kamuidrome.prism import (
    cleanup_from_index,
    find_minecraft_dir,
    get_prism_instances_directory,
)


class ModSide(enum.Enum):
    """
    Enumeration of the possible sides a mod can be on.
    """

    CLIENT = "client"
    SERVER = "server"
    BOTH = "both"


@attr.s(slots=True, kw_only=True, frozen=False)
class InstalledMod:
    """
    Wraps data about a single installed mod.
    """

    #: The name of this mod.
    name: str = attr.ib()

    #: The actual Modrinth ID for this mod.
    project_id: ProjectId = attr.ib()

    #: The pretty, human-readable version for this mod.
    version: str = attr.ib()

    #: The installed version of this mod.
    version_id: VersionId = attr.ib()

    #: The Blake2b hash for this mod. Used as the filename in the cache.
    checksum: str = attr.ib()

    #: If this mod was added explicitly or not (i.e. as a dependency).
    selected: bool = attr.ib()

    #: If this mod is pinned (i.e. won't be automatically updated).
    pinned: bool = attr.ib(default=False)

    #: If true, this is a client-side only mod.
    client_side_only: bool = attr.ib(default=False)


class LocalPack:
    """
    A single, local modpack.
    """

    def __init__(
        self,
        pack_dir: Path,
        metadata: PackMetadata,
        mods: dict[ProjectId, InstalledMod],
    ) -> None:
        self.directory: Path = pack_dir
        self.metadata: PackMetadata = metadata
        self.mods: dict[ProjectId, InstalledMod] = mods

    def _write_index(self):
        """
        Writes out the mod index for this pack.
        """

        mod_index = self.directory / "mods" / "mod-index.json"
        serialised = cattrs.unstructure(self.mods)

        with mod_index.open(mode="w") as f:
            json.dump(serialised, f, indent=4, sort_keys=True)
            f.write(os.linesep)

    def download_and_add_mods(
        self,
        api: ModrinthApi,
        cache: ModCache,
        versions: VersionResult,
        selected_mod: ProjectId | None,
    ) -> None:
        """
        Downloads all of the provided mods and adds them to this pack's index.
        """

        # selected_mod is used for setting InstalledMod's explicit selection parameter.
        # pin is used to explicitly pin the selected mod.

        with Progress() as progress:
            # we do it this way so that they all show up at first, and then
            # the all_mods is at the bottom.

            tasks_by_mod = {
                version.project_id: progress.add_task(
                    description=f"{project.title} {version.version_number}",
                    total=version.primary_file.size,
                )
                for (project, version) in versions
            }

            all_mods = progress.add_task("[green]Downloading mods...", total=len(versions))

            for project, version in versions:
                old_metadata = self.mods.get(project.id)
                exists_already = cache.get_real_filename(version.project_id, version.id) is not None

                current_task = tasks_by_mod[version.project_id]

                if exists_already:
                    print(
                        f"[yellow]skipping[/yellow] "
                        f"[bold white]{project.title}[/bold white] download as it exists already"
                    )

                    progress.remove_task(current_task)
                else:
                    selected_file = version.primary_file

                    with api.get_file(selected_file.url) as resp:
                        for chunk in cache.save_mod_from_response(
                            resp, version.project_id, version.id, selected_file.filename
                        ):
                            progress.update(current_task, advance=chunk)

                    progress.update(current_task, completed=selected_file.size)

                new_checksum = cache.get_file_checksum(version.project_id, version.id)
                selected = selected_mod == version.project_id
                if old_metadata is not None:
                    old_checksum = old_metadata.checksum

                    # obviously, we only validate the checksum if we're downloading the same
                    # version. previously, i forgot to constrain this by version, so it would always
                    # raise a checksum mismatch if the version was different (and the jar file
                    # was obviously different).
                    if old_metadata.version_id == version.id and old_checksum != new_checksum:
                        raise ValueError(
                            "Invalid saved checksum for "
                            f"{project.title} -> {version.version_number}!"
                        )

                    if old_metadata.pinned:
                        print(
                            f"[yellow]not updating[/yellow] "
                            f"[bold white]{project.title}[/bold white] metadata as it is pinned"
                        )
                        progress.update(all_mods, advance=1)
                        continue

                    if not selected:
                        selected = old_metadata.selected

                self.mods[project.id] = InstalledMod(
                    name=project.title,
                    project_id=version.project_id,
                    version=version.version_number,
                    version_id=version.id,
                    checksum=cast(str, cache.get_file_checksum(version.project_id, version.id)),
                    selected=selected,
                    pinned=False,
                    client_side_only=project.server_side == ModSideValue.UNSUPPORTED,
                )

                progress.update(all_mods, advance=1)

        self._write_index()

    def _validate_downloaded_mods(self, cache: ModCache) -> bool:
        """
        Validates downloaded mods to check if they all exist.
        """

        any_error = False

        for project_id, mod in self.mods.items():
            path = cache.get_mod_path(project_id, mod.version_id)
            if not path.exists():
                print(
                    f"[red]missing mod:[/red] "
                    f"[white]{mod.name}[/white] ([white]{mod.version}[/white])"
                )
                any_error = True

        return not any_error

    def _setup_instance_firsttime(
        self,
        instance_path: Path,
    ) -> None:
        """
        Sets up an instance for the first time.
        """

        mods_dir = instance_path / "mods"
        shutil.rmtree(mods_dir, ignore_errors=True)
        mods_dir.mkdir(exist_ok=False, parents=False)

        configs_dir = instance_path / "config"
        shutil.rmtree(configs_dir, ignore_errors=True)

        # path traversal! oh no!
        for dir in self.metadata.include_directories:
            shutil.rmtree(instance_path / dir, ignore_errors=True)

    def deploy_to_directory(
        self,
        cache: ModCache,
        deploy_path: Path,
        localmeta: LocalMetadata | None,
    ) -> int:
        """
        Deploys the symbolic links to the provided directory.
        """

        if not self._validate_downloaded_mods(cache):
            print("[red]unable to validate downloaded mods.[/red] try 'kamuidrome download' first.")
            return 1

        deploy_path.mkdir(exist_ok=True, parents=False)

        index_path = deploy_path / "kamuidrome.json"

        # step 1: make sure there's no stale symlinks around
        if not index_path.exists():
            print("[yellow]no index found[/yellow], cleaning up instance files")
            self._setup_instance_firsttime(deploy_path)
        else:
            print("[yellow]cleaning up symlinks from index...[/yellow]")
            cleanup_from_index(deploy_path, index_path)

        symlink_index: list[str] = []

        def symlink(symlink_file: Path, original: Path, is_dir: bool) -> None:
            symlink_file.symlink_to(original, target_is_directory=is_dir)
            symlink_index.append(str(symlink_file))

        # step 2: symlink the custom directories
        our_base_dir = self.directory.resolve()
        include_directories = ["config", *self.metadata.include_directories]

        for directory in include_directories:
            potential_dir = our_base_dir / directory
            if not potential_dir.exists():
                print(f"[yellow]skipping dir[/yellow] [white]{potential_dir}[/white] (not found)")
                continue

            included_symlink = deploy_path / directory
            if included_symlink.exists():
                print(
                    f"[yellow]removing old, non-symlink dir[/yellow] "
                    f"[white]{included_symlink}[/white]"
                )
                shutil.rmtree(included_symlink)

            symlink(included_symlink, potential_dir, is_dir=True)

            print(f"[green]linked included dir[/green] [white]{included_symlink}[/white]")

        # step 3: symlink found mod jar files
        our_mods_dir = our_base_dir / "mods"
        deployed_mods_dir = deploy_path / "mods"
        for file in our_mods_dir.iterdir():
            if file.suffix != ".jar":
                continue

            # don't overwrite if there's a ``.disabled`` in the target directory.
            target_file = deployed_mods_dir / file.name
            if target_file.with_suffix(".jar.disabled").exists():
                print(f"[yellow]skipping deploying[/yellow] [white]{target_file.name}[/white]")
                continue

            symlink(target_file, file, is_dir=False)

            print(f"[green]linked included mod[/green] [white]{target_file}[/white]")

        # step 4: symlink mods from cache
        for mod in self.mods.values():
            actual_file_location = cache.get_mod_path(mod.project_id, mod.version_id)
            actual_file_name = cache.get_real_filename(mod.project_id, mod.version_id)
            assert actual_file_name, "don't fuck about with me!"

            target_file = deployed_mods_dir / actual_file_name
            if target_file.with_suffix(".jar.disabled").exists():
                print(f"[yellow]skipping deploying[/yellow] [white]{target_file.name}[/white]")
                continue

            symlink(target_file, actual_file_location.resolve(), is_dir=False)

            print(f"[green]linked managed mod[/green] [white]{target_file}[/white]")

        # step 5: symlink local directories
        if localmeta is not None:
            for extra_dir in localmeta.extra_symlinked_dirs:
                our_extra_dir = (our_base_dir / extra_dir).resolve()
                if not our_extra_dir.exists():
                    continue

                their_extra_dir = deploy_path / extra_dir
                symlink(their_extra_dir, our_extra_dir, is_dir=True)
                print(f"[green]linked extra dir[/green] [white]{our_extra_dir}[/white]")

        with index_path.open(mode="w") as f:
            json.dump(symlink_index, f)

        return 0

    def deploy_to_instance(
        self, cache: ModCache, instance_name: str, localmeta: LocalMetadata | None
    ) -> int:
        """
        Deploys a modpack to the specified Prism Launcher instance.
        """

        if not self._validate_downloaded_mods(cache):
            print("[red]unable to validate downloaded mods.[/red] try 'kamuidrome download' first.")
            return 1

        prism_dir = get_prism_instances_directory()
        instance_dir = find_minecraft_dir(prism_dir, instance_name)
        instance_dir = instance_dir.resolve()

        self.deploy_to_directory(cache, instance_dir, localmeta)

        return 0

    def pin(self, mod_name: str) -> int:
        """
        Pins a single mod to its currently installed version.
        """

        try:
            selected_mod = self.mods[ProjectId(mod_name)]
        except KeyError:
            for mod in self.mods.values():
                if mod.name.lower() == mod_name.lower():
                    selected_mod = mod
                    break
            else:
                print(f"[red]unknown mod:[/red] [bold white]{mod_name}[/bold white]")
                return 1

        selected_mod.pinned = True
        self._write_index()

        print(
            f"[green]pinned mod[/green] [bold white]{selected_mod.name}[/bold white] "
            f"to version [bold white]{selected_mod.version}[/bold white]"
        )
        return 0


def find_pack_dir() -> Path | None:
    """
    Finds the pack directory by looking through parent directories.
    """

    # st_dev is the ID of the device that the filesystem for this file resides on.
    # see stat(3type)

    top_fs = Path.cwd().stat().st_dev
    dirs = deque([Path.cwd()])

    while dirs:
        next_dir = dirs.popleft()
        if next_dir.stat().st_dev != top_fs:
            print(
                f"[red]refusing to resolve through different filesystem at[/red] "
                f"[white]{next_dir}[/white]"
            )
            return None

        if (next_dir / "pack.toml").exists():
            return next_dir

        if next_dir.parent == next_dir:  # root directory
            return None

        dirs.append(next_dir.parent)

    raise RuntimeError("unreachable")


def load_local_pack(pack_dir: Path) -> LocalPack:
    """
    Loads a :class:`.LocalPack` from the specified directory.
    """

    with (pack_dir / "pack.toml").open() as f:
        raw_data = toml_load(f)
        pack_meta = cattrs.structure(raw_data, PackMetadata)

    try:
        mods_index_path = pack_dir / "mods" / "mod-index.json"
        with mods_index_path.open() as f:
            raw_selected = json.load(f)
            selected_mods = cattrs.structure(raw_selected, dict[ProjectId, InstalledMod])
    except FileNotFoundError:
        selected_mods: dict[ProjectId, InstalledMod] = {}

    return LocalPack(pack_dir, pack_meta, selected_mods)



import contextlib
import enum
import subprocess
from subprocess import SubprocessError

import attr


class AvailablePackLoader(enum.Enum):
    """
    Enumeration of the supported pack loaders.
    """

    LEGACY_FORGE = "legacyforge"
    FABRIC = "fabric"
    QUILT = "quilt"
    NEOFORGE = "neoforge"


@attr.s(slots=True, kw_only=True)
class PackLoaderInfo:
    """
    Definition for modpack loader information.
    """

    #: The name of the modloader to use.
    type: AvailablePackLoader = attr.ib()

    #: The version of the modloader, if any.
    version: str | None = attr.ib(default=None)

    #: For "legacyforge" and "neoforge", this enables using Fabric mods and treating
    #: Forgified Fabric API as Fabric API for dependency purposes.
    sinytra_compat: bool = attr.ib(default=False)

    #: Hack to work around geckolib issues.
    prefer_fabric_geckolib: bool = attr.ib(default=True)

    def get_modrinth_facets(self, *, pretend_to_be_forge: bool = False) -> list[str]:
        """
        Gets a list of Modrinth facets for the loader information within.
        """

        if self.type == AvailablePackLoader.FABRIC or self.type == AvailablePackLoader.QUILT:
            return [f"categories:{self.type.value}"]

        facets: list[str] = []
        if self.sinytra_compat:
            facets.append("categories:fabric")

        if self.type == AvailablePackLoader.LEGACY_FORGE or pretend_to_be_forge:
            facets.append("categories:forge")

        if self.type == AvailablePackLoader.NEOFORGE:
            facets.append("categories:neoforge")

        return facets

    @property
    def modrinth_facets(self) -> list[str]:
        """
        Gets a list of Modrinth facets for the loader information within.
        """

        return self.get_modrinth_facets(pretend_to_be_forge=False)

    @property
    def mrpack_name(self) -> str:
        """
        Returns the ``mrpack`` dependency name for this loader.
        """

        match self.type:
            case AvailablePackLoader.FABRIC:
                return "fabric-loader"

            case AvailablePackLoader.QUILT:
                return "quilt-loader"

            case AvailablePackLoader.LEGACY_FORGE:
                return "forge"

            case AvailablePackLoader.NEOFORGE:
                return "neoforge"


@attr.s(slots=True, frozen=True, kw_only=True)
class PackMetadata:
    """
    Base definition for a pack file.
    """

    #: Human-friendly name of the pack, used during export.
    name: str = attr.ib()

    #: Human-friendly version of the pack, used during export.
    version: str = attr.ib(default="0.0.0")

    #: If True, dynamic versioning (Git versioning) is enabled.
    dynamic_version: bool = attr.ib(default=False)

    #: The minecraft version for this pack.
    game_version: str = attr.ib()

    #: The additional directories to include.
    include_directories: list[str] = attr.ib()

    loader: PackLoaderInfo = attr.ib()

    def __attrs_post_init__(self):
        if not self.dynamic_version:
            return

        with contextlib.suppress(SubprocessError):
            described = subprocess.check_output("git describe".split(), encoding="utf-8").strip()

            if described.startswith("v"):
                described = described[1:]

            object.__setattr__(self, "version", described)

    def get_available_loaders(self, *, pretend_to_be_forge: bool = False):
        """
        Returns the available modloaders, in priority order.
        """

        match self.loader.type:
            case AvailablePackLoader.FABRIC:
                return ("fabric",)

            case AvailablePackLoader.QUILT:
                return ("quilt", "fabric")

            case AvailablePackLoader.LEGACY_FORGE if self.loader.sinytra_compat:
                return ("forge", "fabric")

            case AvailablePackLoader.LEGACY_FORGE:
                return ("forge",)

            case AvailablePackLoader.NEOFORGE if (
                pretend_to_be_forge and self.loader.sinytra_compat
            ):
                return ("forge", "fabric")

            case AvailablePackLoader.NEOFORGE if self.loader.sinytra_compat:
                return ("neoforge", "fabric")

            case AvailablePackLoader.NEOFORGE:
                return ("neoforge",)

    @property
    def available_loaders(self) -> tuple[str] | tuple[str, str]:
        """
        Returns the available modloaders, in priority order.
        """

        match self.loader.type:
            case AvailablePackLoader.FABRIC:
                return ("fabric",)

            case AvailablePackLoader.QUILT:
                return ("quilt", "fabric")

            case AvailablePackLoader.LEGACY_FORGE if self.loader.sinytra_compat:
                return ("forge", "fabric")

            case AvailablePackLoader.LEGACY_FORGE:
                return ("forge",)

            case AvailablePackLoader.NEOFORGE if self.loader.sinytra_compat:
                return ("neoforge", "fabric")

            case AvailablePackLoader.NEOFORGE:
                return ("neoforge",)


@attr.s(slots=True, kw_only=True)
class LocalMetadata:
    """
    Wrapper for the data within the ``localpack.toml``.
    """

    #: The name of the instance to deploy to automatically.
    instance_name: str = attr.ib()

    #: Extra directories to symlink, but not to include.
    extra_symlinked_dirs: list[str] = attr.ib(factory=list)



from rich.progress import Progress

from kamuidrome.cache import ModCache
from kamuidrome.modrinth.client import ModrinthApi
from kamuidrome.modrinth.models import ProjectId
from kamuidrome.modrinth.utils import (
    VersionResult,
    resolve_dependency_versions,
    resolve_latest_version,
)
from kamuidrome.pack import LocalPack


def download_all_mods(
    pack: LocalPack,
    modrinth: ModrinthApi,
    cache: ModCache,
) -> int:
    """
    Downloads all mods in the index for a specified pack.
    """

    projects = {p.id: p for p in modrinth.get_multiple_projects(list(pack.mods.keys()))}
    versions = modrinth.get_multiple_versions([v.version_id for v in pack.mods.values()])
    all_versions: VersionResult = []

    for ver in versions:
        all_versions.append((projects[ver.project_id], ver))

    pack.download_and_add_mods(modrinth, cache, all_versions, selected_mod=None)

    return 0


def update_all_mods(
    pack: LocalPack,
    modrinth: ModrinthApi,
    cache: ModCache,
) -> int:
    """
    Updates all mods in the index for a specified pack.
    """

    all_versions: VersionResult = []
    deps_seen: set[ProjectId] = set()

    with Progress() as progress:
        task = progress.add_task("Fetching mod info", total=len(pack.mods))

        projects = modrinth.get_multiple_projects(list(pack.mods.keys()))

        for mod in projects:
            latest_version = resolve_latest_version(pack.metadata, modrinth, mod)
            all_versions.append((mod, latest_version))
            all_versions += resolve_dependency_versions(
                pack.metadata, modrinth, latest_version, _seen=deps_seen
            )
            progress.advance(task, 1)

    # de-duplicate downloads
    seen: set[ProjectId] = set()
    to_download: VersionResult = []

    for info, version in all_versions:
        if info.id in seen:
            continue

        seen.add(info.id)

        to_download.append((info, version))

    if __debug__:
        for mod, version in to_download:
            assert (
                mod.id == version.project_id
            ), f"{mod.title} has non-matching version {version.id}"

    pack.download_and_add_mods(modrinth, cache, to_download, selected_mod=None)

    return 0



from collections.abc import Iterable

from rich import box, print
from rich.table import Table
from rich.text import Text

from kamuidrome.pack import InstalledMod, LocalPack


def _apply_table(table: Table, mods: Iterable[InstalledMod]) -> None:
    table.add_column("Name", justify="left")
    table.add_column("Version", justify="center")
    table.add_column("Pinned", justify="right")

    for mod in mods:
        pinned = "No" if not mod.pinned else "Yes"
        pinned_style = "yellow" if not mod.pinned else "green"
        table.add_row(
            Text(mod.name, style="magenta", justify="left"),
            mod.version,
            Text(pinned, justify="right", style=pinned_style),
        )

    print(table)


def list_indexed_mods(pack: LocalPack, include_deps: bool = True) -> int:
    """
    Lists the indexed mods for the current pack.
    """

    if include_deps:
        dependencies = [mod for mod in pack.mods.values() if not mod.selected]
        dep_table = Table(
            title="[italic]Dependency mods[/italic] (not explicitly selected)",
            box=box.SIMPLE,
            min_width=50,
        )
        _apply_table(dep_table, dependencies)
        print()

    selected = [mod for mod in pack.mods.values() if mod.selected]
    selected_table = Table(title="Selected mods", box=box.SIMPLE, min_width=50)
    _apply_table(selected_table, selected)

    return 0



import enum
import typing
from collections.abc import Iterator, Sequence
from typing import Literal, NewType, final, overload

import attr
import cattr
from cattr import Converter, override
from cattrs.gen._consts import AttributeOverride
from whenever import Instant

# Revert this commit when cattrs 24.x drops.
# type ProjectType = Literal["mod", "modpack", "resourcepack", "shader"]
# type ModSide = Literal["unknown", "required", "optional", "unsupported"]


class ModSideValue(enum.Enum):  # noqa: D101
    UNKNOWN = "unknown"
    REQUIRED = "required"
    OPTIONAL = "optional"
    UNSUPPORTED = "unsupported"


# deliberately undocumented because i'm lazy. see the modrinth docs.
# see client.py for why we need some hackery with cattrs overrides.
# duck typing kinda works at least...

ProjectId = NewType("ProjectId", str)
VersionId = NewType("VersionId", str)


@attr.s(frozen=True, kw_only=True)
class ProjectInfoMixin:
    """
    Common fields between :class:`.ProjectInfoFromSearch` and :class:`.ProjectInfoFromProject`.
    """

    @classmethod
    def configure_converter(cls, converter: Converter) -> None:  # noqa: D102
        # some renames are needed because modrinth is really fucking stupid
        # and likes

        for klass in (ProjectInfoFromSearch, ProjectInfoFromProject):
            converter.register_structure_hook(
                klass,
                cattr.gen.make_dict_structure_fn(
                    klass,
                    converter,
                    raw_categories=override(rename="categories"),
                    raw_versions=override(rename="versions"),
                    **klass.get_overrides(),  # type: ignore  # ok, pyright
                ),
            )

    #: The ID of this project.
    id: ProjectId

    #: The project's URL slug.
    slug: str = attr.ib()
    #: The human-readable title for the project.
    title: str = attr.ib()
    #: Project's short description.
    description: str = attr.ib()  # short description

    # really terrible way of doing it, imo.
    client_side: ModSideValue = attr.ib()
    server_side: ModSideValue = attr.ib()

    raw_categories: list[str] = attr.ib()
    raw_versions: list[str] = attr.ib()


@attr.s(slots=True, frozen=True, kw_only=True)
@final
class ProjectInfoFromSearch(ProjectInfoMixin):
    """
    Wrapper type for project information returned from ``/search``.
    """

    @classmethod
    def get_overrides(cls) -> dict[str, AttributeOverride]:  # noqa: D102
        return {
            "id": override(rename="project_id"),
        }

    id: ProjectId = attr.ib()  # base62 encoded, but we treat it as an opaque string

    @property
    def game_versions(self) -> list[str]:
        """
        Gets the game versions this project is for.
        """

        return self.raw_versions


@attr.s(slots=True, frozen=True, kw_only=True)
@final
class ProjectInfoFromProject(ProjectInfoMixin):
    """
    Wrapper type for project information returned from ``/project``.
    """

    @classmethod
    def get_overrides(cls) -> dict[str, AttributeOverride]:  # noqa: D102
        return {}

    id: ProjectId = attr.ib()

    #: The game versions this project is for.
    game_versions: list[str] = attr.ib()


@attr.s(slots=True, frozen=True, kw_only=True)
@final
class ProjectSearchResult(Sequence[ProjectInfoFromSearch]):
    """
    Wrapper type for the result of a project search.
    """

    # wtf is this name
    hits: list[ProjectInfoFromSearch] = attr.ib()
    offset: int = attr.ib()
    limit: int = attr.ib()
    total_hits: int = attr.ib()

    def __bool__(self) -> bool:
        return bool(self.hits)

    @typing.override
    def __iter__(self) -> Iterator[ProjectInfoFromSearch]:
        return iter(self.hits)

    @typing.override
    def __len__(self) -> int:
        return len(self.hits)

    @overload
    def __getitem__(self, index: int) -> ProjectInfoFromSearch: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[ProjectInfoFromSearch]: ...

    @typing.override
    def __getitem__(
        self,
        index: int | slice,
    ) -> ProjectInfoFromSearch | Sequence[ProjectInfoFromSearch]:
        return self.hits[index]


@attr.s(slots=True, frozen=True, kw_only=True)
@final
class ProjectVersionFile:
    """
    A single file for a project version.
    """

    #: The URL for this version.
    url: str = attr.ib()

    #: The real filename for this version.
    filename: str = attr.ib()

    #: If this version is the *primary* file or not (i.e. the main mod file).
    primary: bool = attr.ib(default=False)

    #: The size of the file, in bytes.
    size: int = attr.ib()

    #: Opaque set of hashes in unspecified algorithms.
    hashes: dict[str, str] = attr.ib()


@attr.s(slots=True, frozen=True, kw_only=True)
@final
class ProjectVersionRelation:
    """
    A single relation within a project version.
    """

    #: The ID of the other project version in this relationship.
    version_id: VersionId | None = attr.ib()

    #: The ID of the other project in this relationship.
    project_id: ProjectId = attr.ib()

    #: The type of this relationship.
    dependency_type: Literal["required", "optional", "incompatible", "embedded"] = attr.ib()


@attr.s(slots=True, frozen=True, kw_only=True, hash=False, eq=False)
@final
class ProjectVersion:
    """
    A single version within a project.
    """

    @classmethod
    def configure_converter(cls, converter: Converter) -> None:  # noqa: D102
        converter.register_structure_hook(
            ProjectVersion,
            cattr.gen.make_dict_structure_fn(
                ProjectVersion, converter, relationships=override(rename="dependencies")
            ),
        )

    #: The ID for this version.
    id: VersionId = attr.ib()

    #: The project ID for this version.
    project_id: ProjectId = attr.ib()

    #: The name of this version.
    name: str = attr.ib()

    #: The number for this version.
    version_number: str = attr.ib()

    #: The game versions for this mod.
    game_versions: list[str] = attr.ib()

    #: If this version is featured or not.
    featured: bool = attr.ib()

    #: The list of modloaders this version is for.
    loaders: list[str] = attr.ib()

    #: When this version was uploaded.
    date_published: Instant = attr.ib()

    #: The stability status for this version.
    version_type: Literal["release", "beta", "alpha"] = attr.ib()

    #: The relationships for this version.
    relationships: list[ProjectVersionRelation] = attr.ib(factory=list)

    #: The list of :class:`.ProjectVersionFile` instances for this single version (e.g. main mod
    #: and sources jar, for weirdos who use Modrinth maven).
    files: list[ProjectVersionFile] = attr.ib()

    @typing.override
    def __hash__(self) -> int:
        return hash(self.project_id)

    @typing.override
    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, ProjectVersion):
            return NotImplemented

        return self.project_id == __value.project_id

    def __attrs_post_init__(self) -> None:
        if len(self.files) == 1:
            object.__setattr__(self.files[0], "primary", True)

    @property
    def primary_file(self) -> ProjectVersionFile:
        """
        Gets the primary file for this project.
        """

        for file in self.files:
            if file.primary:
                return file

        raise ValueError("modrinth fucked up, missing primary version")



import json
from collections.abc import Generator, Sequence
from contextlib import contextmanager
from importlib import metadata
from typing import Literal

import cattr
import httpx
from whenever import Instant

from kamuidrome.modrinth.models import (
    ProjectId,
    ProjectInfoFromProject,
    ProjectInfoMixin,
    ProjectSearchResult,
    ProjectVersion,
    VersionId,
)
from kamuidrome.retry import RetryTransport

CONVERTER = cattr.GenConverter()
CONVERTER.forbid_extra_keys = False
CONVERTER.register_structure_hook(Instant, lambda it, _: Instant.parse_rfc3339(it))

ProjectInfoMixin.configure_converter(CONVERTER)
ProjectVersion.configure_converter(CONVERTER)
VERSION = metadata.version("kamuidrome")


# Here's my small Complaining About The Modrinth API section.
#
# the ``/search`` API and the ``/project`` API have both fields named arbitrarily differently
# (``project_id`` in ``/search``, ``id`` in ``project`` - what fucking ID could it be in search
# that isn't the project ID!) *and* fields with the same name but DIFFERENT meanings (``versions``
# in ``/search`` is the *game versions*, whereas ``versions`` in ``/project`` refers to *project*
# versions instead, and it has a ``game_versions`` field instead.)
#
# the ratelimit stuff is blatant "this was made by somebody who wrote a Discord wrapper library".
# i'm ignoring it because I don't care about hitting your ratelimits.
#
# search queries are some horrible json array query parameter syntax that is passed directly
# to meilisearch (lol) so if you fuck it up you get an unhelpful error message from either the
# backend (because fake json parsing failed) or meilisearch complaining about indexes. fun fucking
# fact: when i was writing this today (2024-01-13), i trieed using a query over ``game_version``
# and it didn't work even though the docs said it would. then the site went down, and it started
# working again immediately after. what the hell is up with that?
#
# STOP MAKING ME PUT JSON IN QUERY STRINGS. are they fucking allergic to commas? hmm yes today
# i will urlescape ``["1.19.2"]``. what a bunch of wasted work for no reason.
#
# minor nitpick: the API returns the sha-1 hash (what?) and the sha512 hash which sucks because
# i use blake2b hashes.
#
# more will certainly come as i deal more with the API, but I am already annoyed.


class ModrinthApi:
    """
    Wrapper around the Modrinth API.
    """

    API_VERSION = "v2"

    def __init__(self, client: httpx.Client) -> None:
        self.client = client
        self.client._transport = RetryTransport(self.client._transport)  # type: ignore
        client.base_url = f"https://api.modrinth.com/{self.API_VERSION}"
        client.headers = {
            "user-agent": f"Mozilla/5.0 (kamuidrome/{VERSION}; https://github.com/Fuyukai/kamuidrome) AppleWebKit/537.3 (KHTML, like Packwiz)"  # noqa: E501
        }
        client.timeout = 10
        client.follow_redirects = True

    def get_project_info(self, project_id: str) -> ProjectInfoFromProject:
        """
        Gets info about the specified project.
        """

        resp = self.client.get(f"/project/{project_id}")
        resp.raise_for_status()
        body = resp.json()
        return CONVERTER.structure(body, ProjectInfoFromProject)

    def get_project_versions(
        self,
        project_id: ProjectId,
        *,
        loaders: Sequence[str] | None = None,
        game_versions: list[str] | str | None = None,
    ) -> list[ProjectVersion]:
        """
        Gets the list of versions for a single project.

        Whilst its not documented anywhere, this is returned in descending upload time order.
        """

        params: dict[str, str] = {}
        if loaders:
            params["loaders"] = json.dumps(loaders)

        if game_versions:
            if isinstance(game_versions, str):
                # fucking modrinth. why?
                game_versions = [game_versions]

            params["game_versions"] = json.dumps(game_versions)

        resp = self.client.get(f"/project/{project_id}/version", params=params)
        resp.raise_for_status()
        return CONVERTER.structure(resp.json(), list[ProjectVersion])

    def get_projects_via_search(
        self,
        search_query: str,
        *facets: str | list[str],
        index: Literal["relevance", "downloads", "follows", "newest", "updated"] = "relevance",
        offset: int = 0,
        limit: int = 100,
    ) -> ProjectSearchResult:
        """
        Gets a list of projects via Modrinth search.
        """

        # modrinth uses this really fucking stupid format for searching
        # it's like... nested json?
        query_facets = [[f] if isinstance(f, str) else f for f in facets]
        resp = self.client.get(
            "/search",
            params={
                "facets": json.dumps(query_facets),
                "query": search_query,
                "offset": offset,
                "limit": limit,
            },
        )
        resp.raise_for_status()
        return CONVERTER.structure(resp.json(), ProjectSearchResult)

    def get_multiple_projects(self, projects: list[ProjectId]) -> list[ProjectInfoFromProject]:
        """
        Gets multiple projeects in one request.
        """

        resp = self.client.get("/projects", params={"ids": json.dumps(projects)})
        resp.raise_for_status()

        return CONVERTER.structure(resp.json(), list[ProjectInfoFromProject])

    def get_single_version(self, version: VersionId) -> ProjectVersion:
        """
        Gets a single version for a project.
        """

        resp = self.client.get(f"/version/{version}")
        resp.raise_for_status()

        return CONVERTER.structure(resp.json(), ProjectVersion)

    def get_multiple_versions(self, versions: list[VersionId]) -> list[ProjectVersion]:
        """
        Gets multiple versions in one request.
        """

        resp = self.client.get(
            "/versions",
            params={
                "ids": json.dumps(versions),
            },
        )
        resp.raise_for_status()

        return CONVERTER.structure(resp.json(), list[ProjectVersion])

    @contextmanager
    def get_file(self, url: str) -> Generator[httpx.Response, None, None]:
        """
        Gets a single file by URL, using this client's ``httpx.Client``.
        """

        with self.client.stream("GET", url) as r:
            yield r



import random
from collections.abc import Iterable, Mapping
from time import sleep
from typing import override

import httpx


# https://github.com/encode/httpx/issues/108
class RetryTransport(httpx.BaseTransport):  # noqa: D101
    RETRYABLE_METHODS = frozenset(["HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE"])
    RETRYABLE_STATUS_CODES = frozenset([413, 429, 503, 504])

    MAX_BACKOFF_WAIT = 60

    def __init__(
        self,
        wrapped_transport: httpx.BaseTransport,
        max_attempts: int = 10,
        max_backoff_wait: float = MAX_BACKOFF_WAIT,
        backoff_factor: float = 0.1,
        jitter_ratio: float = 0.1,
        respect_retry_after_header: bool = True,
        retryable_methods: Iterable[str] | None = None,
        retry_status_codes: Iterable[int] | None = None,
    ) -> None:
        self.wrapped_transport = wrapped_transport
        if jitter_ratio < 0 or jitter_ratio > 0.5:
            raise ValueError(f"jitter ratio should be between 0 and 0.5, actual {jitter_ratio}")

        self.max_attempts = max_attempts
        self.backoff_factor = backoff_factor
        self.respect_retry_after_header = respect_retry_after_header
        self.retryable_methods = (
            frozenset(retryable_methods) if retryable_methods else self.RETRYABLE_METHODS
        )
        self.retry_status_codes = (
            frozenset(retry_status_codes) if retry_status_codes else self.RETRYABLE_STATUS_CODES
        )
        self.jitter_ratio = jitter_ratio
        self.max_backoff_wait = max_backoff_wait

    def _calculate_sleep(
        self, attempts_made: int, headers: httpx.Headers | Mapping[str, str]
    ) -> float:
        backoff = self.backoff_factor * (2 ** (attempts_made - 1))
        jitter = (backoff * self.jitter_ratio) * random.choice([1, -1])
        total_backoff = backoff + jitter
        return min(total_backoff, self.max_backoff_wait)

    @override
    def handle_request(self, request: httpx.Request) -> httpx.Response:
        remaining_attempts = self.max_attempts - 1
        attempts_made = 1

        while True:
            try:
                response = self.wrapped_transport.handle_request(request)
            except httpx.TimeoutException:
                continue

            if remaining_attempts < 1 or response.status_code not in self.retry_status_codes:
                return response

            response.close()

            sleep_for = self._calculate_sleep(attempts_made, response.headers)
            sleep(sleep_for)

            response = self.wrapped_transport.handle_request(request)

            attempts_made += 1
            remaining_attempts -= 1



from __future__ import annotations

import json
import re
from pathlib import Path

import platformdirs


def get_prism_subdir(
    base_dir: Path,
    config_text: str,
    key: str,
    default_name: str,
) -> Path:
    """
    Gets the appropriate sub-directory from the Prism Launcher configuration.

    :param base_dir: The base Prism Launcher directory (e.g. ``~/.local/share/PrismLauncher``).
    :param config_text: The contents of the Prism config file.
    :param key: The key to extract from the config text.
    :param default_name: The default name to use if unset, e.g. ``instances``.
    """

    rxp = rf"{key}=(.*)$"
    matched = re.search(rxp, config_text, re.MULTILINE)
    dir_name = Path(default_name if matched is None else matched.group(1))
    if not dir_name.is_absolute():
        dir_name = base_dir / dir_name

    return dir_name


def get_prism_instances_directory() -> Path:
    """
    Gets the location of the Prism Launcher instances directory.
    """

    base_dir = Path(platformdirs.user_data_dir("PrismLauncher"))
    config_file = base_dir / "prismlauncher.cfg"
    return get_prism_subdir(
        base_dir, config_file.read_text(), key="InstanceDir", default_name="instances"
    )


def find_minecraft_dir(instances_dir: Path, instance: str):
    """
    Finds the ``.minecraft`` directory within a Prism Launcher instance directory.
    """

    instance_path = (instances_dir / instance).absolute()

    for path in (".minecraft", "minecraft"):
        if (what := instance_path / path).exists():
            return what

    raise FileNotFoundError(f"Can't find the ``.minecraft`` for {instance_path}")


def cleanup_from_index(instance_path: Path, index_path: Path) -> None:
    """
    Cleans up created symlinks from an index file.
    """

    with index_path.open() as f:
        index: list[str] = json.load(f)

    for fp in index:
        path = Path(fp)

        if not path.exists(follow_symlinks=False):
            # ok?
            continue

        if not path.is_symlink():
            # ok x2?
            continue

        path.unlink()



import hashlib
import json
from collections.abc import Generator
from pathlib import Path

import attr
import cattrs
from httpx import Response

from kamuidrome.modrinth.models import ProjectId, VersionId


@attr.s(slots=True, kw_only=True)
class ModVersionMetadata:
    """
    Contains metadata for a single mod version.
    """

    #: The real filename for this mod.
    real_file_name: str = attr.ib()

    #: The blake2b hash for this mod version.
    blake_hexdigest: str = attr.ib()


class ModCache:
    """
    Stores information related to mods in the mod cache.
    """

    def __init__(self, cache_dir: Path) -> None:
        self.path = cache_dir

        self.path.mkdir(exist_ok=True)

    def get_mod_path(self, id: ProjectId, version_id: VersionId) -> Path:
        """
        Gets the path to a mod ``.jar`` file.

        This path may or may not exist.
        """

        return (self.path / id / version_id).with_suffix(".jar")

    def _get_metadata(self, id: ProjectId) -> dict[str, ModVersionMetadata]:
        path = self.path / id / "metadata.json"
        try:
            content = cattrs.structure(json.loads(path.read_text()), dict[str, ModVersionMetadata])
        except FileNotFoundError:
            content = {}

        return content

    def get_real_filename(self, id: ProjectId, version_id: VersionId) -> str | None:
        """
        Gets the real filename for the provided mod pair, or None if it is an unknown mod/version.
        """

        try:
            return self._get_metadata(id)[version_id].real_file_name
        except (FileNotFoundError, KeyError):
            return None

    def _update_metadata(
        self,
        project_id: ProjectId,
        version_id: VersionId,
        filename: str,
        hash: str,
    ) -> None:
        """
        Updates the metadata for a single mod version file.
        """

        content = self._get_metadata(project_id)
        content[version_id] = ModVersionMetadata(real_file_name=filename, blake_hexdigest=hash)
        with (self.path / project_id / "metadata.json").open(mode="w") as f:
            json.dump(cattrs.unstructure(content), f)

    def get_file_checksum(self, project_id: ProjectId, version_id: VersionId) -> str | None:
        """
        Gets the file checksum for the provided mod, or None if it is an unknown mod/version.
        """

        try:
            return self._get_metadata(project_id)[version_id].blake_hexdigest
        except KeyError:
            return None

    def save_mod_from_response(
        self,
        response: Response,
        project_id: ProjectId,
        version_id: VersionId,
        real_file_name: str,
    ) -> Generator[int, None, None]:
        """
        Downloads a single mod from the provided :class:`.Response`.
        """

        mod_path = self.get_mod_path(project_id, version_id)
        mod_path.parent.mkdir(exist_ok=True)

        summer = hashlib.blake2b()

        with mod_path.open(mode="wb") as f:
            for chunk in response.iter_bytes(16 * 1024):  # 16KiB chunks seems reasonable
                summer.update(chunk)
                f.write(chunk)
                yield len(chunk)

        self._update_metadata(project_id, version_id, real_file_name, summer.hexdigest())



import httpx
from rich import print
from rich.prompt import IntPrompt

from kamuidrome.cache import ModCache
from kamuidrome.meta import AvailablePackLoader
from kamuidrome.modrinth.client import ModrinthApi
from kamuidrome.modrinth.models import ProjectId, ProjectInfoMixin, VersionId
from kamuidrome.modrinth.utils import (
    FABRIC_API_VERSION,
    resolve_dependency_versions,
    resolve_latest_version,
)
from kamuidrome.pack import LocalPack


def _common_from_project_id(
    pack: LocalPack,
    client: ModrinthApi,
    cache: ModCache,
    project_id: ProjectId | ProjectInfoMixin,
) -> int:
    """
    Common code for any path that uses a project ID.
    """

    if isinstance(project_id, ProjectInfoMixin):
        project_info = project_id
    else:
        project_info = client.get_project_info(project_id)

    if (
        pack.metadata.loader.type
        in (AvailablePackLoader.LEGACY_FORGE, AvailablePackLoader.NEOFORGE)
        and project_id == FABRIC_API_VERSION
    ):
        print("[red]error:[/red] cowardly refusing to install Fabric API on a forge instance")
        return 1

    version = resolve_latest_version(pack.metadata, client, project_info)
    all_versions = [
        (project_info, version),
        *resolve_dependency_versions(pack.metadata, client, version),
    ]
    pack.download_and_add_mods(client, cache, all_versions, selected_mod=project_info.id)

    return 0


def add_mod_by_searching(
    pack: LocalPack,
    client: ModrinthApi,
    cache: ModCache,
    query: str,
    always_prompt_selection: bool,
    pretend_to_be_forge: bool,
) -> int:
    """
    Adds a new mod by searching Modrinth.

    If ``always_prompt_selection`` is passed, then the regular name will be ignored. Useful for
    CLI search.
    """

    result = client.get_projects_via_search(
        query,
        f"game_versions:{pack.metadata.game_version}",
        pack.metadata.loader.get_modrinth_facets(pretend_to_be_forge=pretend_to_be_forge),
        "project_type:mod",
        limit=10,
    )

    if not result:
        print("[red]error:[/red] no results found")
        return 1

    matched = result[0]

    # written weirdly to simplify reading the logic rather than trying to unfuck an all().
    if len(result) == 1:
        should_auto_match = True

    elif always_prompt_selection:
        should_auto_match = False

    elif matched.title.lower() == query.lower():
        should_auto_match = True

    else:
        should_auto_match = False

    if not should_auto_match:
        print("[yellow]No exact match found[/yellow], listing possible options")

        for idx, option in enumerate(result):
            print(rf"[white]\[{idx}][/white] - [bold white]{option.title}[/bold white]")

        option = IntPrompt.ask(
            prompt="Pick an option", show_choices=False, choices=list(map(str, range(len(result))))
        )

        matched = result[option]
    else:
        print(f"[green]successful match[/green]: {matched.title} / {matched.id}")

    return _common_from_project_id(pack, client, cache, matched.id)


def add_mod_by_project_id(
    pack: LocalPack, client: ModrinthApi, cache: ModCache, project_id: ProjectId
) -> int:
    """
    Adds a new mod by project ID.
    """

    try:
        result = client.get_project_info(project_id)
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 404:
            print("[red]error:[/red] no such project found")
            return 1

        raise

    return _common_from_project_id(pack, client, cache, result)


def add_mod_by_version_id(
    pack: LocalPack, client: ModrinthApi, cache: ModCache, version_id: VersionId
):
    """
    Adds a new mod by an explicit version ID.
    """

    try:
        found_version = client.get_single_version(version_id)
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 404:
            print("[red]error:[/red] no such version found")
            return 1

        raise

    project_info = client.get_project_info(found_version.project_id)

    if pack.metadata.game_version not in found_version.game_versions:
        print(
            f"[red]error:[/red] [white]{project_info.title} {found_version.name}[/white] "
            f"does not support [white]{pack.metadata.game_version}[/white]"
        )
        return 1

    if not (set(pack.metadata.available_loaders) | set(found_version.loaders)):
        print(
            f"[red]error:[/red] [white]{project_info.title} {found_version.name}[/white] "
            f"does not support any of {pack.metadata.available_loaders}"
        )
        return 1

    all_versions = [
        (project_info, found_version),
        *resolve_dependency_versions(pack.metadata, client, found_version),
    ]
    pack.download_and_add_mods(client, cache, all_versions, selected_mod=project_info.id)

    return 0



from collections.abc import Sequence

from rich import print

from kamuidrome.exc import NoAvailableVersionsError
from kamuidrome.meta import AvailablePackLoader, PackMetadata
from kamuidrome.modrinth.client import ModrinthApi
from kamuidrome.modrinth.models import (
    ProjectId,
    ProjectInfoFromProject,
    ProjectInfoMixin,
    ProjectVersion,
)

type VersionResult = Sequence[tuple[ProjectInfoMixin, ProjectVersion]]

# Hardcoded Modrinth project IDs used to swap out dependencies easily.

FABRIC_API_VERSION = ProjectId("P7dR8mSH")
FORGIFIED_API_VERSION = ProjectId("Aqlf1Shp")
MODMENU_API_ID = ProjectId("mOgUt4GM")
CONNECTOR_EXTRAS_ID = ProjectId("FYpiwiBR")
FORGE_CONFIG_PORT_ID = ProjectId("ohNO6lps")
GECKOLIB_API_ID = ProjectId("8BmcQJ2H")

DEPENDENCY_SWAPS: dict[ProjectId, ProjectId] = {
    FABRIC_API_VERSION: FORGIFIED_API_VERSION,
    MODMENU_API_ID: CONNECTOR_EXTRAS_ID,
    FORGE_CONFIG_PORT_ID: CONNECTOR_EXTRAS_ID,
}


def _do_resolve_latest_version(
    pack: PackMetadata,
    modrinth: ModrinthApi,
    info: ProjectInfoMixin | ProjectId,
    available_loaders: Sequence[str],
    allow_unstable: bool = False,
) -> ProjectVersion:
    """
    Resolves the latest matching version for the specified mod.

    If ``allow_unstable`` is False, then the most recent *stable* version is chosen; otherwise,
    unstable (alpha and beta) versions will be picked.
    """

    if not isinstance(info, ProjectInfoMixin):
        info = modrinth.get_project_info(info)

    project_id = info.id

    versions = modrinth.get_project_versions(
        project_id=project_id, loaders=available_loaders, game_versions=pack.game_version
    )

    # a bit of trickiness with multiple loader scenarios.
    # some mods are dual fabric-forge (or dual quilt-fabric); and due to the way modrinth orders
    # them, the *secondary* loader might be first.
    # so what we do is we will *always* select *any* version for our preferred loader first,
    # and only fall back to our alternative loader if we fail to select any version for the
    # primary loader.

    versions = sorted(versions, key=lambda it: it.date_published, reverse=True)

    selected_version: ProjectVersion | None = None
    secondary_version: ProjectVersion | None = None

    primary_loader: str
    secondary_loader: str | None = None

    if project_id == ProjectId("8BmcQJ2H") and (
        pack.loader.sinytra_compat and pack.loader.prefer_fabric_geckolib
    ):
        print("[bold yellow]forcing geckolib onto fabric...[/bold yellow]")
        primary_loader = "fabric"
    else:
        primary_loader = available_loaders[0]

        if len(available_loaders) == 2:
            secondary_loader = available_loaders[1]

    # we do a single-pass strategy here using two local variables.
    for version in versions:
        title = info.title

        if primary_loader in version.loaders:
            # a bit of trickiness here; if we find a release version for our preferred loader,
            # that's the version we want to pick *always*.
            # if we find a non-release version for our non-preferred loader, we only select it
            # if we haven't selected a release before, then continue on to try and find a release
            # version.

            if version.version_type == "release" or allow_unstable:
                selected_version = version
                print(
                    f"[green]saving primary version[/green] "
                    f"[bold white]{version.version_number}[/bold white] "
                    f"for [bold white]{title}[/bold white]"
                )
                break
            elif (  # noqa: RET508
                selected_version is None
            ):
                selected_version = version

                print(
                    f"[italic yellow]saving fallback primary (non-release) version[/italic yellow] "
                    f"[bold white]{version.version_number}[/bold white] ({version.loaders}) "
                    f"for [bold white]{title}[/bold white]"
                )
                continue

        # only assign if ``secondary_version`` is None, because we don't want to replace it with
        # an older version for no reason.
        if all(
            (
                secondary_loader is not None,
                secondary_loader in version.loaders,
                secondary_version is None,
            )
        ):
            print(
                f"[italic yellow]saving fallback secondary (off-loader) version[/italic yellow] "
                f"[bold white]{version.version_number}[/bold white] ({version.loaders}) "
                f"for [bold white]{title}[/bold white]"
            )
            # never break here; there might be a primary version coming *after* this one.
            secondary_version = version
        else:
            print(
                f"[red]rejected version[/red] [bold white]{version.version_number}[/bold white] "
                f"for [bold white]{title}[/bold white]"
            )

    if selected_version is not None:
        title = info.title if isinstance(info, ProjectInfoFromProject) else selected_version.name
        num = selected_version.version_number
        print(
            "[green]selected version[/green] "
            f"[bold white]{num} ({selected_version.loaders})[/bold white] "
            f"for [bold white]{title}[/bold white]"
        )
        return selected_version

    if secondary_version is not None:
        title = info.title if isinstance(info, ProjectInfoFromProject) else secondary_version.name
        num = secondary_version.version_number
        print(
            "[green]selected secondary version[/green] "
            f"[bold white]{num} ({secondary_version.loaders})[/bold white] "
            f"for [bold white]{title}[/bold white]"
        )
        return secondary_version

    raise NoAvailableVersionsError(
        f"Couldn't find an appropriate version for {info.title}; "
        f"no valid versions found for {available_loaders} on {pack.game_version}"
    )


def resolve_latest_version(
    pack: PackMetadata,
    modrinth: ModrinthApi,
    info: ProjectInfoMixin | ProjectId,
    allow_unstable: bool = False,
) -> ProjectVersion:
    """
    Resolves the latest matching version for the specified mod.

    If ``allow_unstable`` is False, then the most recent *stable* version is chosen; otherwise,
    unstable (alpha and beta) versions will be picked.

    This is a wrapper function that will force a retry
    """

    retry_for_forge = "neoforge" in pack.available_loaders and pack.game_version == "1.20.1"

    try:
        ver = _do_resolve_latest_version(
            pack, modrinth, info, pack.available_loaders, allow_unstable=allow_unstable
        )

        if "fabric" not in ver.loaders or not retry_for_forge:
            # yeah, retry for legacyforge.
            # sorry modrinth.
            return ver

        print("[yellow]warning: retrying selection for native forge version[/yellow]")

    except NoAvailableVersionsError:
        if pack.game_version != "1.20.1" or pack.loader.type != AvailablePackLoader.NEOFORGE:
            raise

    # force a search for legacyforge instead
    return _do_resolve_latest_version(
        pack,
        modrinth,
        info,
        available_loaders=pack.get_available_loaders(pretend_to_be_forge=True),
        allow_unstable=allow_unstable,
    )


def transform_dependencies(pack: PackMetadata, projects: list[ProjectId]) -> list[ProjectId]:
    """
    Transforms dependencies depending on certain pack metadata.
    """

    # TODO: make this more generic.

    if not pack.loader.sinytra_compat:
        return projects

    return [DEPENDENCY_SWAPS.get(i, i) for i in projects]


def get_set_of_dependencies(pack: PackMetadata, version: ProjectVersion) -> list[ProjectId]:
    """
    Gets the appropriate set of required dependencies given the provided :class:`.ProjectVersion`.
    """

    resolved: list[ProjectId] = []
    for dep in version.relationships:
        if dep.dependency_type != "required":
            continue

        if pack.loader.sinytra_compat:
            resolved.append(DEPENDENCY_SWAPS.get(dep.project_id, dep.project_id))
        else:
            resolved.append(dep.project_id)

    return resolved


def resolve_dependency_versions(
    pack: PackMetadata,
    modrinth: ModrinthApi,
    selected_version: ProjectVersion,
    _seen: set[ProjectId] | None = None,
) -> VersionResult:
    """
    Recursively resolves the dependency versions of the provided selected version.
    """

    dependencies = get_set_of_dependencies(pack, selected_version)

    seen: set[ProjectId] = _seen if _seen is not None else set()

    resolved: set[ProjectVersion] = set()

    while True:
        if not dependencies:
            break

        next_dependencies: list[ProjectId] = []

        for project in dependencies:
            if project in seen:
                continue

            seen.add(project)
            selected_version = resolve_latest_version(pack, modrinth, project)
            resolved.add(selected_version)
            next_dependencies += get_set_of_dependencies(pack, selected_version)

        dependencies = next_dependencies

    project_infos = modrinth.get_multiple_projects([i.project_id for i in resolved])
    zipped: VersionResult = []

    # unlikely to be too slow unless a projeect has like 100 deps.
    for proj in project_infos:
        for version in resolved:
            if version.project_id == proj.id:
                zipped.append((proj, version))
                break

    assert len(zipped) == len(resolved)
    return zipped



# fun fucking fact: modrinth changed from using the relatively clean (albeit, still evil) gitbook
# to a fucking SUPPORT WEBSITE TEMPLATE for their *technical* documentation. what the fuck is
# wrong with them?
# luckily, the internet archive does not forget, so here's a nice direct link to the fullwidth
# page: https://web.archive.org/web/20231009172349/https://docs.modrinth.com/modpacks/format
#
# the documentation itself is pretty poor, so here's some Unofficial Documentation that
# is a bit easier to read.
#
# ``modpack.mrpack`` files are zip files with a bunch of metadata stuffed inside.
# the actual folders and such inside the zip are simply unpacked into the ``.minecraft`` directory.
# there's a bunch of metadata stored within the ``modrinth.index.json`` (which seemingly isn't
# minified?) which contains basic information about the pack and externally downloaded mods.
#
# in their seemingly infinite wisdom, modrinth decided the mod references within would use
# URLs to modrintn's CDN. rather than using version IDs. because the format is fucking insanely
# generic for no reason.

import json
import shutil
import subprocess
import tempfile
from io import StringIO
from pathlib import Path
from typing import Any

import httpx
import more_itertools
from rich import print

from kamuidrome.meta import AvailablePackLoader
from kamuidrome.modrinth.client import ModrinthApi
from kamuidrome.modrinth.models import VersionId
from kamuidrome.pack import LocalPack


def _parse_forge_version(
    body: list[dict[str, Any]],
    game_version: str,
) -> str:
    """
    Gets the latest Forge version from Prism's metadata.
    """

    for entry in body:
        version = entry["version"]
        requires = entry["requires"]

        for i in requires:
            if i["uid"] == "net.minecraft" and i["equals"] == game_version:
                return version

    raise ValueError("Couldn't find a valid (neo|legacy)forge version!")


def select_latest_loader_version(
    client: httpx.Client, game_version: str, loader: AvailablePackLoader
) -> str:
    """
    Gets the latest version for the specified modloader.
    """

    match loader:
        case AvailablePackLoader.FABRIC:
            result = client.get("https://meta.fabricmc.net/v2/versions/loader")
            result.raise_for_status()
            body: list[dict[str, Any]] = result.json()

            return more_itertools.first(filter(lambda it: it["stable"] is True, body))["version"]

        case AvailablePackLoader.QUILT:
            result = client.get("https://meta.quiltmc.org/v3/versions/loader")
            result.raise_for_status()
            body: list[dict[str, Any]] = result.json()

            return body[0]["version"]

        case AvailablePackLoader.LEGACY_FORGE:
            result = client.get("https://meta.prismlauncher.org/v1/net.minecraftforge")
            result.raise_for_status()
            return _parse_forge_version(result.json()["versions"], game_version)

        case AvailablePackLoader.NEOFORGE:
            result = client.get("https://meta.prismlauncher.org/v1/net.neoforged")
            result.raise_for_status()
            return _parse_forge_version(result.json()["versions"], game_version)


# https://stackoverflow.com/a/69375880/15026456
def make_archive(source: Path, destination: Path) -> None:
    """
    Creates a new archive from the given directory.
    """

    base_name = destination.parent / destination.stem
    shutil.make_archive(
        str(base_name),
        "zip",
        root_dir=str(source),
    )


def create_mrpack(
    pack: LocalPack,
    api: ModrinthApi,
    output: Path,
    ci_mode: bool = False,
    server_only: bool = False,
) -> Path:
    """
    Creates a new ``mrpack`` file from the given pack.
    """

    # initially i tried writing directly to the zipfilee, but python ``zipfile`` is an evil module
    # that sucks (i miss java.nio).
    # so instead we just make a temporary directory and use ``shutil.make_archive``.

    with tempfile.TemporaryDirectory() as dir:
        tmpdir_path = Path(dir)
        dot_minecraft = tmpdir_path / "overrides"
        dot_minecraft.mkdir(exist_ok=False, parents=False)

        version_ids: list[VersionId] = []
        for mod in pack.mods.values():
            if server_only and mod.client_side_only:
                print(f"[yellow]not exporting {mod.name}[/yellow]")
                continue

            print(f"[yellow] definitely exporting {mod.name} ({mod.version_id})")
            version_ids.append(mod.version_id)

        versions = api.get_multiple_versions(version_ids)
        files = [version.primary_file for version in versions]

        loader_version = pack.metadata.loader.version
        if loader_version is None:
            loader_version = select_latest_loader_version(
                api.client,
                pack.metadata.game_version,
                pack.metadata.loader.type,
            )

        print(
            "[green]selected loader version[/green] "
            f"[white]{pack.metadata.loader.mrpack_name} {loader_version}[/white]"
        )

        files_struct: list[dict[str, Any]] = []
        index = {
            "formatVersion": 1,
            "game": "minecraft",
            "versionId": pack.metadata.version,
            "name": pack.metadata.name,
            "dependencies": {
                "minecraft": pack.metadata.game_version,
                pack.metadata.loader.mrpack_name: loader_version,
            },
            "files": files_struct,
        }

        # versions in this case means the indexed mods.
        for version_file in files:
            print("adding version file", version_file.filename)
            body = {
                "path": f"mods/{version_file.filename}",
                "hashes": version_file.hashes,
                "downloads": [version_file.url],
                "fileSize": version_file.size,
            }
            files_struct.append(body)

        with (tmpdir_path / "modrinth.index.json").open(mode="w") as f:
            json.dump(index, f, sort_keys=True)

        with (tmpdir_path / "overrides" / "kamuidrome.metadata").open(mode="w") as f:
            f.reconfigure(line_buffering=True)

            # Line 1: Pack name
            f.write(pack.metadata.name)
            f.write("\n")
            # Line 2: Pack version
            f.write(pack.metadata.version)
            f.write("\n")

            # Line 3: (Commit hash, Branch name, Commit message)
            try:
                git_buf = StringIO()
                current_commit = subprocess.check_output(
                    "git rev-parse HEAD".split(), encoding="utf-8"
                ).strip()
                current_branch = subprocess.check_output(
                    "git rev-parse --abbrev-ref HEAD".split(), encoding="utf-8"
                ).strip()
                commit_message = subprocess.check_output(
                    "git show-branch --no-name HEAD".split(), encoding="utf-8"
                ).strip()

                git_buf.write(current_commit)
                git_buf.write(" ")
                git_buf.write(current_branch)
                git_buf.write(" ")
                git_buf.write(commit_message)
                git_line = git_buf.getvalue()
            except subprocess.SubprocessError:
                git_line = ""

            f.write(git_line)
            f.write("\n")

        directories = ["config", "mods", *pack.metadata.include_directories]

        for to_copy_dir in directories:
            real_dir = (pack.directory / to_copy_dir).resolve()
            shutil.copytree(real_dir, dot_minecraft / to_copy_dir)

        if ci_mode:
            shutil.copytree(tmpdir_path, output, dirs_exist_ok=True)
        else:
            actual_file = output.with_suffix(".zip")
            make_archive(tmpdir_path, output)
            actual_file.rename(output)

        print(f"[green]written output to [/green] [white]{output}[/white] ({ci_mode=})")

    return output



import argparse
import contextlib
import pprint
import sys
from pathlib import Path

import cattrs
import httpx
import platformdirs
import tomlkit
from rich import print

from kamuidrome.cache import ModCache
from kamuidrome.cli.add import add_mod_by_project_id, add_mod_by_searching, add_mod_by_version_id
from kamuidrome.cli.init import interactively_create_pack
from kamuidrome.cli.list import list_indexed_mods
from kamuidrome.cli.update import download_all_mods, update_all_mods
from kamuidrome.meta import LocalMetadata
from kamuidrome.modrinth.client import ModrinthApi
from kamuidrome.modrinth.models import ProjectId, VersionId
from kamuidrome.modrinth.mrpack import create_mrpack
from kamuidrome.pack import find_pack_dir, load_local_pack


def main() -> int:
    """
    Main entrypoint.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--pack-dir",
        help="The directory to use (defaults to CWD)",
        default=None,
        type=Path,
    )
    parser.add_argument(
        "--cache-dir",
        help="The cache directory to use (defaults to ~/.cache)",
        default=Path(platformdirs.user_cache_dir("kamuidrone")),
        type=Path,
    )

    subcommands = parser.add_subparsers(
        title="Subcommands",
        description="Pack management actions",
        dest="subcommand",
        required=True,
    )

    init_group = subcommands.add_parser("init", help="Initialise a new modpack interactively")
    init_group.add_argument(
        "--git", help="Create a Git repository and .gitignore", action="store_true", default=False
    )

    add_mod = subcommands.add_parser("add", help="Adds a new mod")
    add_mod.add_argument(
        "--always-select",
        help="Always shows a selection prompt for search",
        action="store_true",
        default=False,
    )
    add_mod.add_argument(
        "--pretend-to-be-forge",
        action="store_true",
        help="Does a search pretending to be Legacyforge instead of Neoforge (1.20.1) only",
        default=False,
    )
    add_group = add_mod.add_mutually_exclusive_group(required=True)
    add_group.add_argument(
        "-s", "--search", help="Adds a mod by searching for the specified argument", default=None
    )
    add_group.add_argument(
        "-p", "--project-id", help="Adds a mod using the specified project ID", default=None
    )
    add_group.add_argument(
        "-V", "--version-id", help="Adds a mod using the specified version ID", default=None
    )

    deploy_group = subcommands.add_parser("deploy", help="Deploys a modpack")
    deploy_group_args = deploy_group.add_mutually_exclusive_group(required=False)
    deploy_group_args.add_argument(
        "-i", "--instance", help="The name of the Prism instance to deploy to", default=None
    )
    deploy_group_args.add_argument(
        "-d", "--directory", help="The path of the directory to deploy to", default=None, type=Path
    )

    pin_group = subcommands.add_parser(name="pin", help="Pins a mod version to the current version")
    pin_group.add_argument("MOD", nargs="+", help="The mod name or ID to pin")

    subcommands.add_parser(name="list", help="List indexed mods")

    subcommands.add_parser(name="download", help="Downloads all mods in the index")
    subcommands.add_parser(name="update", help="Updates all mods and dependenciess in the index")

    export_group = subcommands.add_parser("export", help="Exports pack as an mrpack file")
    export_group.add_argument(
        "FILENAME", help="The name of the file or directory to write", nargs="?", default=None
    )
    export_group.add_argument(
        "--ci-mode",
        action="store_true",
        default=False,
        help="Outputs an unpacked mrpack instead of a fully packed one",
    )
    export_group.add_argument(
        "--server-only", action="store_true", default=False, help="Only outputs server-side mods"
    )

    args = parser.parse_args()
    cache_dir: Path = args.cache_dir
    cache = ModCache(cache_dir=cache_dir)

    subcommand: str = args.subcommand

    pack_dir: Path | None = args.pack_dir
    if subcommand == "init":
        if pack_dir is None:
            pack_dir = Path.cwd()

        return interactively_create_pack(pack_dir, with_git=args.git)

    if pack_dir is None:
        pack_dir = find_pack_dir()

        if pack_dir is None:
            print("[red]couldn't find a valid pack.toml[/red]")
            return 1

    pack = load_local_pack(pack_dir)

    with httpx.Client() as client:
        api = ModrinthApi(client)

        if subcommand == "add":
            if args.pretend_to_be_forge and pack.metadata.game_version != "1.20.1":
                parser.error("can't pretend to be forge on version != 1.20.1")

            search_query: str | None = args.search
            if search_query is not None:
                return add_mod_by_searching(
                    pack=pack,
                    client=api,
                    cache=cache,
                    query=search_query,
                    always_prompt_selection=args.always_select,
                    pretend_to_be_forge=args.pretend_to_be_forge,
                )

            project_id: str | None = args.project_id
            if project_id is not None:
                return add_mod_by_project_id(pack, api, cache, ProjectId(project_id))

            version_id: str = args.version_id
            add_mod_by_version_id(pack, api, cache, VersionId(version_id))

        elif subcommand == "deploy":
            instance_name: str | None = args.instance
            folder_name: Path | None = args.directory

            local_metadata: LocalMetadata | None = None

            with contextlib.suppress(FileNotFoundError, KeyError):
                localpack = pack.directory / "localpack.toml"
                data = tomlkit.loads(localpack.read_text())
                local_metadata = cattrs.structure(data, LocalMetadata)

            if folder_name is not None:
                return pack.deploy_to_directory(cache, folder_name, local_metadata)

            if instance_name is None:
                if local_metadata is None:
                    parser.error(
                        "expected either a folder name, an instance name, "
                        "or for the instance name to be set in localpack.toml"
                    )

                instance_name = local_metadata.instance_name

            return pack.deploy_to_instance(cache, instance_name, local_metadata)

        elif subcommand == "pin":
            return pack.pin(" ".join(args.MOD))

        elif subcommand == "download":
            return download_all_mods(pack, api, cache)

        elif subcommand == "update":
            return update_all_mods(pack, api, cache)

        elif subcommand == "list":
            return list_indexed_mods(pack)

        elif subcommand == "export":
            export_name: str | None = args.FILENAME
            ci_mode: bool = args.ci_mode
            server_only: bool = args.server_only

            if export_name is None:
                if ci_mode:
                    parser.error("must provide a directory name in ci mode")
                else:
                    export_path = (Path.cwd() / pack.metadata.name).with_suffix(".mrpack")
            elif not ci_mode:
                export_path = Path(export_name).with_suffix(".mrpack")
            else:
                export_path = Path(export_name)
                export_path.mkdir(exist_ok=True, parents=True)

            create_mrpack(pack, api, export_path, ci_mode=ci_mode, server_only=server_only)
            return 0

    return 0


if __name__ == "__main__":

    def _run():
        try:
            sys.exit(main())
        except httpx.HTTPStatusError as e:
            pprint.pprint(e.response.json())

    _run()



# flake8: noqa: F841

import os
import subprocess
from pathlib import Path
from typing import cast, override

import tomlkit
from rich.prompt import Confirm, InvalidResponse, Prompt, PromptBase

from kamuidrome.meta import AvailablePackLoader


class NonEmptyPrompt(PromptBase[str]):
    """
    A prompt that makes sure the entry is non-empty.
    """

    @override
    def process_response(self, value: str) -> str:
        if not value:
            raise InvalidResponse("[prompt.invalid]Enter a non-empty value")

        return super().process_response(value)


def interactively_create_pack(
    path: Path,
    with_git: bool,
) -> int:
    """
    Creates a new modpack interactively.
    """

    name = NonEmptyPrompt.ask("[yellow]Pack name")
    raw_minecraft_version = NonEmptyPrompt.ask("[yellow]Minecraft version")
    mc_version: tuple[int, int] = cast(
        tuple[int, int], tuple(map(int, raw_minecraft_version.split(".")[1:]))
    )
    assert len(mc_version) == 2, "enter the right version idiot!"

    version = Prompt.ask("[yellow]Pack version", default="0.1.0")

    if mc_version < (20, 0):
        loader_choices = ["legacyforge", "fabric", "quilt"]
    else:
        loader_choices = ["legacyforge", "neoforge", "fabric", "quilt"]

    loader = AvailablePackLoader(Prompt.ask("[yellow]Modloader", choices=loader_choices))
    loader_version: str | None = Prompt.ask("[yellow]Modloader version (leave empty for auto)")
    if not loader_version:
        loader_version = None

    if mc_version >= (20, 0) and loader in (
        AvailablePackLoader.LEGACY_FORGE,
        AvailablePackLoader.NEOFORGE,
    ):
        enable_sinytra = Confirm.ask("[yellow]Enable Sinytra compatibility", default=True)
    else:
        enable_sinytra = False

    if all((enable_sinytra, loader == AvailablePackLoader.LEGACY_FORGE, mc_version == (20, 1))):
        # hardcoded lol.
        loader_version = "47.1.3"

    pack_toml = tomlkit.document()
    pack_toml.add(tomlkit.comment("This pack.toml was generated by ``kamuidrome init``!"))
    pack_toml.add(tomlkit.nl())

    pack_toml.add(tomlkit.comment("Human-friendly name of the pack, used during export."))
    pack_toml["name"] = name
    pack_toml.add(tomlkit.comment("Human-friendly version of the pack, used during export."))
    pack_toml["version"] = version
    pack_toml.add(tomlkit.comment("The minecraft version for this pack."))
    pack_toml["game_version"] = raw_minecraft_version
    pack_toml.add(
        tomlkit.comment(
            "A list of additional directories to include when building a pack or "
            "generating symlinks."
        )
    )
    pack_toml["include_directories"] = []
    pack_toml.add(tomlkit.nl())

    loader_toml = tomlkit.document()
    loader_toml.add(tomlkit.comment("The type of modloader this pack runs."))
    loader_toml["type"] = loader.value

    if loader_version is not None:
        loader_toml.add(
            tomlkit.comment(
                "The loader version is normally automatically selected but it can be overridden here."  # noqa: E501
            )
        )
        loader_toml["version"] = loader_version

    loader_toml.add(
        tomlkit.comment(
            'For "legacyforge" and "neoforge", this enables using Fabric mods and treating Forgified Fabric API'  # noqa: E501
        )
    )
    loader_toml.add(tomlkit.comment("as Fabric API for dependency purposes."))
    loader_toml["sinytra_compat"] = enable_sinytra

    pack_toml["loader"] = loader_toml

    # not 100% why comments aren't preserved in the nested table, but whatever.
    with (path / "pack.toml").open(mode="w") as f:
        tomlkit.dump(pack_toml, f)  # type: ignore

    (path / "config").mkdir(exist_ok=True)
    (path / "mods").mkdir(exist_ok=True)
    (path / "mods" / "mod-index.json").write_text(r"{}" + os.linesep)

    if with_git:
        subprocess.check_call(["git", "init", str(path.resolve())])
        (path / "config" / ".gitkeep").write_text("")

    return 0
